# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy

from cicada.communicator import SocketCommunicator
from cicada.additive import AdditiveProtocolSuite

def main(communicator):
    protocol = AdditiveProtocolSuite(communicator)
    a = numpy.arange(24).reshape((2,3,4))

    ashare = protocol.share(src=0, secret=protocol.encoder.encode(a), shape=a.shape)

    print("revealed:", protocol.encoder.decode(protocol.reveal(ashare)))
    arshare = protocol.reshare(operand=ashare)
    print("revealed reshare:", protocol.encoder.decode(protocol.reveal(arshare)))

SocketCommunicator.run(world_size=3, fn=main)
